local config = require("config")
local math2d = require("math2d")

local function on_chunk_generated(event)
  local surface = event.surface
  local left_top = event.area.left_top
  local right_bottom = event.area.right_bottom

  generate_team_divider(surface, left_top, right_bottom)
  generate_ribbon_edge(surface, left_top, right_bottom)
  generate_start_area(surface, left_top, right_bottom)
  -- clear_biters(surface, left_top, right_bottom)
end

function generate_team_divider(surface, left_top, right_bottom)
  local divider_distance = config.map.divider_chunks_to_center * 32

  if right_bottom.y > divider_distance or left_top.y < -divider_distance then return end

  fill_tiles_with("out-of-map", surface, left_top, right_bottom)
end

function generate_ribbon_edge(surface, left_top, right_bottom)
  local edge_distance = (config.map.divider_chunks_to_center + config.map.team_ribbon_chunk_width) * 32

  if left_top.y < edge_distance and right_bottom.y > -edge_distance and right_bottom.x > (-config.map.left_edge_chunks_to_center * 32) then return end

  fill_tiles_with("out-of-map", surface, left_top, right_bottom)
end

function fill_tiles_with(tile_name, surface, left_top, right_bottom)
  local tiles = {}
  local i = 1

  for y = left_top.y, right_bottom.y-1 do
    for x = left_top.x, right_bottom.x-1 do
      tiles[i] = {name=tile_name, position={x, y}}
      i = i + 1
    end
  end

  surface.set_tiles(tiles)
end

function generate_start_area(surface, left_top, right_bottom)
  local sa_cfg = config.start_area
  -- don't do anything if generated chunk is outside of start area boundary
  if   left_top.x < sa_cfg.left_top_chunk.x * 32 --right cutoff
    or right_bottom.x > sa_cfg.right_bottom_chunk.x * 32 --left cutoff
    or right_bottom.y > sa_cfg.left_top_chunk.y * 32 --top cutoff
    or left_top.y < -sa_cfg.left_top_chunk.y * 32 --bottom cutoff
    or (left_top.y < sa_cfg.right_bottom_chunk.y * 32 and right_bottom.y > -sa_cfg.right_bottom_chunk.y * 32) -- middle cutoff
  then return end

  fill_tiles_with("grass-1", surface, left_top, right_bottom)
  fill_tiles_with("stone-path", surface, left_top, right_bottom)

  -- sign is used to detect which half of map chunk generator is currently working on
  -- it simplifies condition by checking only against one half of the map
  local sign = sign(left_top.y)
  local chunk_bounding_box = {left_top=left_top, right_bottom=right_bottom}

  -- for all ores => if ore center in chunk box -> generate ore
  for entity_name, opts in pairs(sa_cfg.resources) do
    local center = {x=opts.center.x, y=opts.center.y * sign}
    if math2d.bounding_box.contains_point(chunk_bounding_box, center) then
      local ore_bounding_box = math2d.bounding_box.create_from_centre(center, opts.width, opts.height)
      fill_resources_with(surface, entity_name, ore_bounding_box, opts.amount, opts.step or 1)
    end
  end

  -- starting water
  local water_center = {x=sa_cfg.water.center.x, y=sa_cfg.water.center.y * sign}
  if math2d.bounding_box.contains_point(chunk_bounding_box, water_center) then
    local water_bounding_box = math2d.bounding_box.create_from_centre(water_center, sa_cfg.water.width, sa_cfg.water.height)
    fill_tiles_with("water", surface, water_bounding_box.left_top, water_bounding_box.right_bottom)
  end

  -- remove cliffs
  for _, cliff in pairs(surface.find_entities_filtered{type="cliff", area=chunk_bounding_box}) do
    cliff.destroy()
  end


end

function fill_resources_with(surface, entity_name, bounding_box, amount, step)
  for y = bounding_box.left_top.y, bounding_box.right_bottom.y, step do
    for x = bounding_box.left_top.x, bounding_box.right_bottom.x, step do
      local opts = {
        ["name"] = entity_name,
        ["position"] = {x, y},
        ["amount"] = amount
      }
      surface.create_entity(opts)
    end
  end

end

function clear_biters(surface, left_top, right_bottom)
  if math.abs(left_top.x) > config.map.clear_biters_chunks * 32 then return end
  local area = {left_top=left_top, right_bottom=right_bottom}

  for _, entity in pairs(surface.find_entities_filtered{{area = area}, force = "enemy"}) do
    entity.destroy()
  end

end

function sign(x)
  return x>0 and 1 or x<0 and -1 or 0
end

local map = {}

map.events =
{
  [defines.events.on_chunk_generated] = on_chunk_generated,
}


return map
